import collections
import sys
from contextlib import contextmanager

import logbook
from orderedset import OrderedSet

from ..._compat import OrderedDict, iteritems, itervalues, reraise
from ...ctx import context as slash_context
from ...exception_handling import handling_exceptions
from ...exceptions import CyclicFixtureDependency, UnresolvedFixtureStore, UnknownFixtures, InvalidFixtureName
from ...utils.python import get_arguments
from ..variation_factory import VariationFactory
from ..test import is_valid_test_name
from .active_fixture import ActiveFixture
from .fixture import Fixture
from .namespace import Namespace
from .parameters import Parametrization, iter_parametrization_fixtures
from .utils import (get_real_fixture_name_from_argument, get_scope_by_name,
                    nofixtures)

_logger = logbook.Logger(__name__)

class FixtureStore(object):

    def __init__(self):
        super(FixtureStore, self).__init__()
        self._namespaces = [Namespace(self)]
        self._unresolved_fixture_ids = set()
        self._fixtures_by_id = {}
        self._fixtures_by_fixture_info = {}
        self._active_fixtures_by_scope = collections.defaultdict(OrderedDict)
        self._computing = set()
        self._all_needed_parametrization_ids_by_fixture_id = {}
        self._known_fixture_ids = collections.defaultdict(dict) # maps fixture ids to known combinations

    def get_active_fixture(self, fixture):
        return self._active_fixtures_by_scope[fixture.info.scope].get(fixture.info.id)

    def _compute_id(self, variation, p):
        if isinstance(p, Parametrization):
            return variation.param_value_indices[p.info.id]
        combination = frozenset((f.info.id, self._compute_id(variation, f))
                                for f in self.iter_all_needed_fixture_objects(p))
        known = self._known_fixture_ids[p.info.id]
        return known.setdefault(combination, len(known))

    def iter_all_needed_fixture_objects(self, fixtureobj):
        for fid in self.get_all_needed_fixture_ids(fixtureobj):
            yield self.get_fixture_by_id(fid)

    def iter_active_fixtures(self):
        for _, fixtures in self._active_fixtures_by_scope.items():
            for f in fixtures.values():
                yield f

    def call_with_fixtures(self, test_func, namespace, trigger_test_start=False, trigger_test_end=False):

        if not nofixtures.is_marked(test_func):
            fixture_names = self.get_required_fixture_names(test_func)
            kwargs = self.get_fixture_dict(fixture_names, namespace)
        else:
            kwargs = {}

        if trigger_test_start:
            for fixture in self.iter_active_fixtures():
                fixture.call_test_start()

        returned = test_func(**kwargs)

        if trigger_test_end:
            for fixture in self.iter_active_fixtures():
                fixture.call_test_end()

        return returned

    def get_required_fixture_names(self, test_func):
        """Returns a list of fixture names needed by test_func.

        Each element returned is either a string or a tuple of (required_name, real_name)
        """
        skip_names = {name for name, _ in iter_parametrization_fixtures(test_func)}
        returned = []
        for argument in get_arguments(test_func):
            if argument.name in skip_names:
                continue
            real_name = get_real_fixture_name_from_argument(argument)
            if real_name == argument.name:
                returned.append(real_name)
            else:
                returned.append((argument.name, real_name))
        return returned

    def get_required_fixture_objects(self, test_func, namespace):
        names = self.get_required_fixture_names(test_func)
        assert isinstance(names, list)
        return set(itervalues(self.get_fixture_dict(names, namespace=namespace, get_values=False)))

    def resolve_name(self, parameter_name, start_point, namespace=None):

        if namespace is None:
            namespace = self.get_current_namespace()

        parts = parameter_name.split('.')[::-1]

        if not parts:
            raise UnknownFixtures(parameter_name)

        while parts:
            current_name = parts.pop()
            param_fixtures = dict(iter_parametrization_fixtures(start_point))
            if current_name in param_fixtures:
                if parts: # we cannot decend further than a parameter
                    raise UnknownFixtures(parameter_name)
                start_point = param_fixtures[current_name]
            else:
                start_point = self.get_fixture_by_name(current_name, namespace=namespace)
                namespace = start_point.namespace
        return start_point

    def __iter__(self):
        return itervalues(self._fixtures_by_id)

    def push_namespace(self):
        self._namespaces.append(Namespace(self, parent=self._namespaces[-1]))

    def pop_namespace(self):
        return self._namespaces.pop(-1)

    @contextmanager
    def new_namespace_context(self):
        self.push_namespace()
        try:
            yield
        finally:
            self.pop_namespace()

    def get_current_namespace(self):
        return self._namespaces[-1]

    def get_all_needed_fixture_ids(self, fixtureobj):
        if self._unresolved_fixture_ids:
            raise UnresolvedFixtureStore()
        if isinstance(fixtureobj, Parametrization):
            return frozenset([fixtureobj.info.id])
        returned = self._all_needed_parametrization_ids_by_fixture_id.get(fixtureobj.info.id)
        if returned is None:
            returned = self._compute_all_needed_parametrization_ids(fixtureobj)
            self._all_needed_parametrization_ids_by_fixture_id[fixtureobj.info.id] = returned
        return returned

    def iter_autouse_fixtures_in_namespace(self, namespace=None):
        if namespace is None:
            namespace = self.get_current_namespace()
        for fixture in namespace.iter_fixtures():
            if fixture.info.autouse:
                yield fixture

    def activate_autouse_fixtures_in_namespace(self, namespace):
        for fixture in self.iter_autouse_fixtures_in_namespace(namespace):
            _ = self.get_fixture_value(fixture)

    def _compute_all_needed_parametrization_ids(self, fixtureobj):
        stack = [(fixtureobj.info.id, [fixtureobj.info.id], set([fixtureobj.info.id]))]
        returned = OrderedSet()
        while stack:
            fixture_id, path, visited = stack.pop()
            if fixture_id in self._all_needed_parametrization_ids_by_fixture_id:
                returned.update(self._all_needed_parametrization_ids_by_fixture_id[fixture_id])
                continue
            fixture = self._fixtures_by_id[fixture_id]
            if fixture.parametrization_ids:
                assert isinstance(fixture.parametrization_ids, OrderedSet)
                returned.update(fixture.parametrization_ids)
            if fixture.keyword_arguments:
                for needed in itervalues(fixture.keyword_arguments):
                    if needed.is_parameter():
                        continue
                    needed_id = needed.info.id
                    if needed_id in visited:
                        self._raise_cyclic_dependency_error(fixtureobj, path, needed_id)
                    stack.append((needed_id, path + [needed_id], visited | set([needed_id])))
        return returned

    def _raise_cyclic_dependency_error(self, fixtureobj, path, new_id):
        raise CyclicFixtureDependency(
            'Cyclic fixture dependency detected in {}: {}'.format(
                fixtureobj.info.func.__code__.co_filename,
                ' -> '.join(self._fixtures_by_id[f_id].info.name
                            for f_id in path + [new_id])))

    def push_scope(self, scope):
        scope = get_scope_by_name(scope)

    def pop_scope(self, scope): # pylint: disable=unused-argument
        if slash_context.result is not None and slash_context.result.is_interrupted():
            return
        scope = get_scope_by_name(scope)
        for s, active_fixtures in iteritems(self._active_fixtures_by_scope):
            if s <= scope:
                for active_fixture in reversed(list(active_fixtures.values())):
                    with handling_exceptions(swallow=True):
                        self._deactivate_fixture(active_fixture.fixture)
                assert not active_fixtures

    def ensure_known_parametrization(self, parametrization):
        if parametrization.info.id not in self._fixtures_by_id:
            self._fixtures_by_id[parametrization.info.id] = parametrization

    def add_fixtures_from_dict(self, d):
        for thing in itervalues(d):
            fixture_info = getattr(thing, '__slash_fixture__', None)
            if fixture_info is None:
                continue
            assert self.get_current_namespace() is self._namespaces[-1]
            fixture_info = self.add_fixture(thing).__slash_fixture__
            self.get_current_namespace().add_name(
                fixture_info.name, fixture_info.id)


    def add_fixture(self, fixture_func):
        fixture_info = fixture_func.__slash_fixture__
        existing_fixture = self._fixtures_by_id.get(fixture_info.id)
        if existing_fixture is not None:
            return existing_fixture.fixture_func
        if is_valid_test_name(fixture_info.name):
            raise InvalidFixtureName('Invalid fixture name: {.name}'.format(fixture_info))
        fixture_object = Fixture(self, fixture_func)
        current_namespace = self._namespaces[-1]
        current_namespace.add_name(fixture_info.name, fixture_info.id)
        self.register_fixture_id(fixture_object)
        return fixture_func

    def register_fixture_id(self, f):
        if f.info.id in self._fixtures_by_id:
            return
        self._fixtures_by_id[f.info.id] = f
        self._unresolved_fixture_ids.add(f.info.id)

    def get_fixture_by_name(self, name, namespace=None):
        if namespace is None:
            namespace = self._namespaces[-1]
        return namespace.get_fixture_by_name(name)

    def get_fixture_by_argument(self, arg):
        return self.get_fixture_by_name(get_real_fixture_name_from_argument(arg))

    def get_fixture_by_id(self, fixture_id):
        return self._fixtures_by_id[fixture_id]

    def get_fixture_dict(self, fixture_names, namespace=None, get_values=True, skip_names=frozenset()):
        returned = {}

        if namespace is None:
            namespace = self.get_current_namespace()

        for element in fixture_names:
            if isinstance(element, tuple):
                required_name, real_name = element
            else:
                required_name = real_name = element

            if required_name in skip_names:
                continue
            if element == 'this':
                continue
            fixture = namespace.get_fixture_by_name(real_name)
            if get_values:
                fixture = self.get_fixture_value(fixture, name=required_name)
            returned[required_name] = fixture
        return returned

    def get_fixture_value(self, fixture, name=None):
        if name is None:
            name = fixture.info.name

        value = self._compute_fixture_value(name, fixture)
        return value

    def get_value(self, variation, parameter_or_fixture):
        fixture_id = parameter_or_fixture.info.id

        fixtureobj = self.get_fixture_by_id(parameter_or_fixture.info.id)
        if isinstance(fixtureobj, Parametrization):
            value = parameter_or_fixture.get_value_by_index(variation.param_value_indices[fixture_id])
        else:
            value = self.get_fixture_value(parameter_or_fixture)

        return value

    def iter_parametrization_variations(self, fixture_ids=(), funcs=(), methods=()):

        if self._unresolved_fixture_ids:
            raise UnresolvedFixtureStore()

        variation_factory = VariationFactory(self)
        for fixture_id in fixture_ids:
            variation_factory.add_needed_fixture_id(fixture_id)

        for func in funcs:
            variation_factory.add_needed_fixtures_from_function(func)

        for method in methods:
            variation_factory.add_needed_fixtures_from_method(method)

        return variation_factory.iter_variations()

    def _compute_fixture_value(self, name, fixture, relative_name=None):
        if relative_name is None:
            relative_name = name

        assert not fixture.is_parameter()

        if fixture.info.id in self._computing:
            raise CyclicFixtureDependency(
                'Fixture {!r} is a part of a dependency cycle!'.format(name))

        active_fixture = self.get_active_fixture(fixture)

        if active_fixture is not None:
            return active_fixture.value

        self._computing.add(fixture.info.id)
        try:
            fixture_value = self._call_fixture(fixture, relative_name=relative_name)
        except:
            exc_info = sys.exc_info()
            self._deactivate_fixture(fixture)
            reraise(*exc_info)
        finally:
            self._computing.discard(fixture.info.id)

        return fixture_value

    def _call_fixture(self, fixture, relative_name):
        assert relative_name
        active_fixture = ActiveFixture(fixture)

        kwargs = {}

        if fixture.keyword_arguments is None:
            raise UnresolvedFixtureStore('Fixture {} is unresolved!'.format(fixture.info.name))

        for required_name, needed_fixture in iteritems(fixture.keyword_arguments):
            if needed_fixture.is_parameter():
                continue
            kwargs[required_name] = self._compute_fixture_value(
                required_name, needed_fixture,
                relative_name='{} -> {}'.format(relative_name, required_name))


        assert fixture.info.id not in self._active_fixtures_by_scope[fixture.info.scope]
        self._active_fixtures_by_scope[fixture.info.scope][fixture.info.id] = active_fixture
        prev_context_fixture = slash_context.fixture
        slash_context.fixture = active_fixture
        try:
            returned = active_fixture.value = fixture.get_value(kwargs, active_fixture)
        finally:
            slash_context.fixture = prev_context_fixture
        _logger.trace(' -- {} = {!r}', relative_name, returned)
        return returned

    def _deactivate_fixture(self, fixture):
        # in most cases it will be the last active fixture in its scope
        active = self._active_fixtures_by_scope[fixture.info.scope].pop(fixture.info.id, None)
        if active is not None:
            active.do_cleanups()

    def resolve(self):
        while self._unresolved_fixture_ids:
            fixture = self._fixtures_by_id[self._unresolved_fixture_ids.pop()]
            fixture.resolve(self)



from __future__ import absolute_import

import inspect
import linecache
import os
import sys
import traceback
import types

from vintage import deprecated

from .._compat import PY2
from .. import context
from .python import get_underlying_func


_MAX_VARIABLE_VALUE_LENGTH = 100

_FILTERED_MEMBER_TYPES = [types.MethodType, types.FunctionType, type]
if PY2:
    _FILTERED_MEMBER_TYPES.append(types.UnboundMethodType) # pylint: disable=no-member
    _FILTERED_MEMBER_TYPES.append(types.ClassType) # pylint: disable=no-member
_FILTERED_MEMBER_TYPES = tuple(_FILTERED_MEMBER_TYPES)

_ALLOWED_ATTRIBUTE_TYPES = [int, str, float]
if PY2:
    _ALLOWED_ATTRIBUTE_TYPES.append(long) # pylint: disable=undefined-variable
_ALLOWED_ATTRIBUTE_TYPES = tuple(_ALLOWED_ATTRIBUTE_TYPES)


def get_traceback_string(exc_info=None):
    if exc_info is None:
        exc_info = sys.exc_info()
    exc_type, exc_value, exc_tb = exc_info  # pylint: disable=unpacking-non-sequence
    return "".join(traceback.format_exception(exc_type, exc_value, exc_tb))


def distill_traceback(tb, frame_correction=0, **kw):
    frames = _get_tb_frames(tb)
    if frame_correction:
        frames = frames[:len(frames) - frame_correction]
    return _distill_frames(frames, **kw)


def distill_call_stack(frame_correction=0, **kw):
    frames = _get_sys_trace_frames()
    if frame_correction:
        frames = frames[:len(frames) - frame_correction + 1]
    return _distill_frames(frames, **kw)


def _distill_frames(frames):
    returned = DistilledTraceback()
    for frame in frames:
        if isinstance(frame, tuple):
            frame, lineno = frame
        else:
            lineno = None
        if _is_frame_and_below_muted(frame):
            break
        if not _is_frame_muted(frame):
            returned.frames.append(DistilledFrame(frame, lineno))
    return returned


def _get_tb_frames(tb):
    returned = []
    while tb is not None:
        returned.append((tb.tb_frame, tb.tb_lineno))
        tb = tb.tb_next
    return returned

def _get_sys_trace_frames():
    return [f[0] for f in reversed(inspect.stack()[:-1])]

def _is_frame_muted(frame):
    try:
        return _deduce_frame_function(frame) in _MUTED_LOCATIONS
    finally:
        del frame


def _deduce_frame_function(frame):
    try:
        frame_module = _deduce_frame_module(frame)
        if frame_module is None:
            return None

        return (frame_module, frame.f_code.co_name)
    finally:
        del frame


def _deduce_frame_module(frame):
    return frame.f_globals.get("__name__")


_MUTED_LOCATIONS = set([
    ("slash.core.function_test", "run"),
    ("slash.core.test", "run"),
    ("slash.exception_handling", "handling_exceptions"),
    ("slash.core.fixtures.fixture_store", "call_with_fixtures"),
    ("slash.frontend.main", "__main__"),
    ("slash.frontend.main", "main"),
    ("slash.frontend.main", "main_entry_point"),
    ("slash.frontend.slash_run", "slash_run"),
    ("slash.runner", "run_tests"),
    ("slash.runner", "_run_single_test"),
    ("slash.core.cleanup_manager", "call_cleanups"),
    ("slash.core.cleanup_manager", "__call__"),
])


def _is_frame_and_below_muted(frame):
    return frame.f_globals.get("__name__") in _MUTE_BELOW_MODULES

_MUTE_BELOW_MODULES = set([
    "slash.assertions"
])


class DistilledTraceback(object):

    def __init__(self):
        super(DistilledTraceback, self).__init__()
        self.frames = []

    def to_list(self):
        return [frame.to_dict() for frame in self.frames]

    @property
    def cause(self):
        if self.frames:
            return self.frames[-1]

    def __repr__(self):
        return self.to_string()

    def to_string(self, include_vars=False):
        return '\n'.join(frame.to_string(include_vars=include_vars) for frame in self.frames)


class DistilledFrame(object):

    def __init__(self, frame, lineno=None):
        super(DistilledFrame, self).__init__()
        self.python_frame = frame
        self.filename = os.path.abspath(frame.f_code.co_filename)
        if lineno is None:
            lineno = frame.f_lineno
        self.lineno = lineno
        self.func_name = frame.f_code.co_name
        self._locals = self._capture_locals(frame)
        self._globals = self._capture_globals(frame)
        self.code_line = linecache.getline(self.filename, self.lineno).rstrip()
        self.code_string = "".join(
            linecache.getline(self.filename, lineno)
            for lineno in range(frame.f_code.co_firstlineno, self.lineno + 1)) or None
        if context.test is not None:
            test_function = get_underlying_func(context.test.get_test_function())
            self._is_in_test_code = frame.f_code is test_function.__code__
        else:
            self._is_in_test_code = False

    @property
    @deprecated(since='1.5.0')
    def locals(self):
        return self._locals

    @property
    @deprecated(since='1.5.0')
    def globals(self):
        return self._globals

    def forget_python_frame(self):
        self.python_frame = None

    def is_in_test_code(self):
        return self._is_in_test_code

    def to_dict(self):
        serialized = {}
        for attr in ['filename', 'lineno', 'func_name', 'code_line', 'code_string']:
            serialized[attr] = getattr(self, attr)
        serialized['globals'] = self._globals
        serialized['locals'] = self._locals
        serialized['is_in_test_code'] = self._is_in_test_code
        return serialized

    def _capture_globals(self, frame):
        used_globals = set(frame.f_code.co_names)
        return dict((global_name, {"value": _safe_repr(value)})
                    for global_name, value in frame.f_globals.items()
                    if global_name in used_globals and self._is_global_included(value))

    def _is_global_included(self, g):
        if isinstance(g, (types.FunctionType, types.MethodType, types.ModuleType, type)):
            return False
        return True

    def _capture_locals(self, frame):
        return dict((local_name, {"value": _safe_repr(local_value)})
                    for key, value in frame.f_locals.items()
                    if "@" not in key
                    for local_name, local_value in self._unwrap_local(key, value))

    def _unwrap_local(self, local_name, local_value):
        yield local_name, local_value
        if local_name != 'self':
            return

        for attr, value in iter_distilled_object_attributes(local_value):
            yield 'self.{}'.format(attr), value

    def __repr__(self):
        return self.to_string()

    def to_string(self, include_vars=False):
        returned = '  {0.filename}, line {0.lineno}:\n'.format(self)
        returned += '    {.code_line}'.format(self)
        if include_vars and self.python_frame is not None:
            for name, value in _unwrap_self_locals(self.python_frame.f_locals.items()):
                returned += '\n\t- {}: {}'.format(name, _safe_repr(value, truncate=False))
            returned += '\n'
        return returned


def _unwrap_self_locals(local_pairs):
    for name, value in local_pairs:
        yield name, value
        if name == 'self':
            for attr_name, attr_value in getattr(value, '__dict__', {}).items():
                yield 'self.{}'.format(attr_name), attr_value


def iter_distilled_object_attributes(obj):
    try:
        obj_dict = getattr(obj, '__dict__', {})
    except Exception:       # pylint: disable=broad-except
        obj_dict = {}

    for attr in obj_dict:
        if attr.startswith('__') and attr.endswith('__'):
            continue
        try:
            value = getattr(obj, attr)
        except Exception:   # pylint: disable=broad-except
            continue
        if isinstance(value, _FILTERED_MEMBER_TYPES):
            continue
        yield attr, value


def distill_object_attributes(obj, truncate=True):

    return {attr: value if isinstance(value, _ALLOWED_ATTRIBUTE_TYPES) else _safe_repr(value, truncate=truncate)
            for attr, value in iter_distilled_object_attributes(obj)}


def _safe_repr(value, truncate=True):
    try:
        returned = repr(value)
    except Exception:  # pylint: disable=broad-except
        return "[Unprintable {!r} object]".format(type(value).__name__)

    if truncate and len(returned) > _MAX_VARIABLE_VALUE_LENGTH:
        returned = returned[:_MAX_VARIABLE_VALUE_LENGTH - 3] + '...'
    return returned



# pylint: disable=unused-argument,redefined-outer-name
import functools
import os

import logbook
import brotli
import gzip
import gossip
import pytest
import slash

from .utils import run_tests_assert_success, run_tests_in_session, TestCase
from .utils.suite_writer import Suite

def test_console_format(suite, suite_test, config_override, tmpdir):
    config_override('log.format', 'file: {record.message}')
    config_override('log.console_format', 'console: {record.message}')
    config_override('log.root', str(tmpdir))
    suite_test.append_line('slash.logger.error("message here")')
    summary = suite.run(additional_args=['-vvv'])

    assert 'console: message here' in summary.get_console_output()

    [result] = summary.get_all_results_for_test(suite_test)
    with open(result.get_log_path()) as f:
        assert 'file: message here' in f.read()


def test_last_session_symlinks(files_dir, links_dir, session):

    test_log_file = files_dir.join(
        session.id, list(session.results.iter_test_results())[-1].test_metadata.id, "debug.log")
    assert test_log_file.check()
    session_log_file = files_dir.join(session.id, "session.log")
    assert session_log_file.check()

    assert links_dir.join("last-session").readlink() == session_log_file
    assert links_dir.join("last-session-dir").readlink() == session_log_file.dirname
    assert links_dir.join("last-test").readlink() == test_log_file


def test_global_result_get_log_path(files_dir, suite):
    summary = suite.run()
    assert summary.session.results.global_result.get_log_path() is not None
    assert summary.session.results.global_result.get_log_path().startswith(str(files_dir))

def _decompress(input_file_name, use_gzip=True):
    if use_gzip:
        with gzip.open(input_file_name, 'rb') as in_f:
            return in_f.read().decode()
    else:
        with open(input_file_name, 'rb') as in_f:
            return brotli.decompress(in_f.read()).decode()

@pytest.mark.parametrize('compression_enabled', [True, False])
@pytest.mark.parametrize('compression_method', ['gzip', 'brotli'])
@pytest.mark.parametrize('use_rotating_raw_file', [True, False])
def test_logs_compression(files_dir, suite, config_override, compression_enabled, compression_method, use_rotating_raw_file):
    config_override("log.compression.enabled", compression_enabled)
    config_override("log.compression.algorithm", compression_method)
    config_override("log.compression.use_rotating_raw_file", use_rotating_raw_file)
    summary = suite.run()
    session_log_path = summary.session.results.global_result.get_log_path()

    if compression_enabled:
        #validate file names
        if compression_method is "gzip":
            session_log_path.endswith("gz")
        else:
            assert session_log_path.endswith(".br")

        assert summary.session.results.global_result.get_log_path().startswith(str(files_dir))

        raw_file_name = session_log_path[:session_log_path.rfind(".")]
        if use_rotating_raw_file:
            assert os.path.exists(raw_file_name)
        else:
            assert not os.path.exists(raw_file_name)

        #validate compressing successfully
        if use_rotating_raw_file:
            decompressed_logs = _decompress(session_log_path, compression_method == "gzip")
            with open(raw_file_name, 'r') as raw_file:
                assert decompressed_logs.endswith(raw_file.read())
    else:
        assert session_log_path.endswith(".log")

def test_compressing_to_unified_file(files_dir, suite, config_override):
    config_override("log.compression.enabled", True)
    config_override("log.subpath", slash.config.root.log.session_subpath)
    config_override("log.compression.use_rotating_raw_file", True)
    config_override("log.compression.algorithm", "gzip")
    summary = suite.run()
    session_log_path = summary.session.results.global_result.get_log_path()

    raw_file_name = session_log_path[:session_log_path.rfind(".")]
    assert os.path.exists(raw_file_name)

    #validate compressing successfully
    decompressed_logs = _decompress(session_log_path, use_gzip=True)
    with open(raw_file_name, 'r') as raw_file:
        assert decompressed_logs.endswith(raw_file.read())

def test_log_file_colorize(files_dir, config_override, suite, suite_test):
    config_override('log.colorize', True)
    suite_test.append_line('slash.logger.notice("hey")')
    summary = suite.run()
    logfiles = [
        summary.session.results.global_result.get_log_path(),
        summary.get_all_results_for_test(suite_test)[0].get_log_path(),
    ]
    for logfile in logfiles:
        with open(logfile, 'rb') as f:
            log_data = f.read()

        assert b'\x1b[' in log_data


@pytest.mark.parametrize('level', ['info', 'notice', 'warning'])
def test_console_truncation_does_not_truncate_files(files_dir, suite, suite_test, config_override, level):
    assert slash.config.root.log.truncate_console_lines

    long_string = 'a' * 1000
    suite_test.append_line('slash.logger.{level}({msg!r})'.format(msg=long_string, level=level))
    summary = suite.run()
    [result] = summary.get_all_results_for_test(suite_test)
    with open(result.get_log_path()) as logfile:
        logfile_data = logfile.read()
        assert long_string in logfile_data


@pytest.mark.parametrize('symlink_name', ['last_session_symlink', 'last_session_dir_symlink', 'last_failed_symlink'])
def test_log_symlinks_without_root_path(suite, config_override, symlink_name):
    config_override('log.{}'.format(symlink_name), 'some/subdir')
    assert suite.run().ok()


def test_last_test_not_overriden_by_stop_on_error(links_dir, suite):
    failed_test = suite[4]
    failed_test.when_run.fail()
    # we stop on error...
    for test in suite[5:]:
        test.expect_not_run()
    summary = suite.run(additional_args=['-x'])

    [failed_result] = summary.get_all_results_for_test(failed_test)

    for link_name in ('last-test', 'last-failed'):
        assert links_dir.join(link_name).readlink() == failed_result.get_log_path()


def test_last_test_delete_log_file(links_dir, suite, suite_test):
    os.makedirs(str(links_dir))
    temp_file = os.path.abspath(str(links_dir.join('somepath')))
    with open(temp_file, 'w'):
        pass
    os.symlink(temp_file, str(links_dir.join('last-test')))
    os.unlink(temp_file)

    assert not os.path.exists(str(links_dir.join('last-test')))
    assert os.path.islink(str(links_dir.join('last-test')))

    summary = suite.run()
    with open(summary.session.results.global_result.get_log_path()) as f:
        assert 'OSError: ' not in f.read()
    # assert links_dir.join('last-test').readlink() == list(summary.session.results)[-1].get_log_path()


def test_result_log_links(files_dir, session):

    for result in session.results.iter_test_results():
        assert result.get_log_path() is not None
        assert result.get_log_path().startswith(str(files_dir))


def test_last_failed(suite, links_dir):
    suite[-5].when_run.fail()
    last_failed = suite[-2]
    last_failed.when_run.fail()
    summary = suite.run()

    [result] = summary.get_all_results_for_test(last_failed)
    fail_log = result.get_log_path()
    assert os.path.isfile(fail_log)
    assert links_dir.join('last-failed').readlink() == fail_log


def test_errors_log_for_test(suite, suite_test, errors_log_path, logs_dir):
    suite_test.when_run.fail()
    res = suite.run()[suite_test]
    with errors_log_path.open() as f:
        lines = [l for l in f.read().splitlines() if 'NOTICE' not in l]
        error_line = lines[0]
    assert 'Error added' in error_line
    with open(res.get_log_path()) as f:
        lines = f.read().splitlines()
        assert error_line in lines

@pytest.mark.parametrize('should_keep_failed_tests', [True, False])
def test_logs_deletion(suite, suite_test, errors_log_path, logs_dir, config_override, should_keep_failed_tests):
    config_override('log.cleanup.enabled', True)
    config_override('log.cleanup.keep_failed', should_keep_failed_tests)

    suite_test.when_run.fail()
    summary = suite.run()
    remaining_files = []
    logs_files_dir = os.path.join(str(logs_dir), 'files')
    for dirname, _, files in os.walk(str(logs_dir)):
        for filename in files:
            f = os.path.join(dirname, filename)
            if os.path.isfile(f) and not os.path.islink(f):
                remaining_files.append(f)
    if should_keep_failed_tests:
        for file_path in [summary[suite_test].get_log_path(), summary.session.results.global_result.get_log_path()]:
            assert file_path in remaining_files
        assert os.listdir(logs_files_dir)
    else:
        assert len(remaining_files) == 0   # pylint: disable=len-as-condition
        assert not os.listdir(logs_files_dir)

def test_errors_log_for_session(suite, errors_log_path, request, logs_dir):
    @gossip.register('slash.session_start')
    def on_session_start():
        try:
            1/0                     # pylint: disable=pointless-statement
        except ZeroDivisionError:
            slash.add_error()

    request.addfinalizer(on_session_start.gossip.unregister)
    results = suite.run(expect_session_errors=True).session.results
    assert len(results.global_result.get_errors()) == 1
    with errors_log_path.open() as f:
        lines = [l for l in f.read().splitlines() if 'NOTICE' not in l]
        assert 'Error added' in lines[0]

    with open(results.global_result.get_log_path()) as f:
        assert lines[0] in f.read().splitlines()


@pytest.mark.parametrize('level', ['info', 'error'])
@pytest.mark.parametrize('core_log_level', [logbook.WARNING, logbook.TRACE])
def test_filtering_slash_logs(files_dir, config_override, level, core_log_level):
    config_override('log.core_log_level', core_log_level)
    assert slash.config.root.log.truncate_console_lines

    suite = Suite()
    suite_test = suite.add_test(type='function')
    long_string = 'a' * 1000
    suite_test.append_line('slash.logger.{level}({msg!r})'.format(msg=long_string, level=level))
    summary = suite.run()
    result = summary.get_all_results_for_test(suite_test)[0]

    with open(result.get_log_path()) as logfile:
        logfile_data = logfile.read()
        assert long_string in logfile_data

    debug_core_string = 'slash.loader: Checking'
    debug_bypass_string = 'slash.runner: #'
    with open(summary.session.results.global_result.get_log_path()) as logfile:
        logfile_data = logfile.read()
        assert debug_bypass_string in logfile_data
        if core_log_level == logbook.TRACE:
            assert debug_core_string in logfile_data
        else:
            assert debug_core_string not in logfile_data

def test_log_names_with_timestamps(files_dir, config_override, suite):
    config_override('log.core_log_level', logbook.TRACE)
    config_override('log.session_subpath', '{timestamp:%Y-%m-%d-%H%M%S}/session.log')
    config_override('log.subpath', '{timestamp:%Y-%m-%d-%H%M%S}/test.log')
    suite.run()


@pytest.mark.parametrize('raw_values', [True, False])
def test_show_raw_param_values(suite_builder, raw_values, config_override, unified_session_log_path):
    config_override('log.show_raw_param_values', raw_values)


    # pylint: disable=unused-variable
    @suite_builder.first_file.add_code
    def __code__():
        import slash # pylint: disable=redefined-outer-name, reimported


        @slash.parametrize('param', [object()])
        def test_1(param):
            pass

    res = suite_builder.build().run().assert_success(1).slash_app.session.results
    with unified_session_log_path.open() as f:
        log_text = f.read()


    raw_log = 'param=<object object at 0x'

    if raw_values:
        assert raw_log in log_text
    else:
        assert 'test_1(param=param0)' in log_text
        assert raw_log not in log_text


################################################################################
## Fixtures

@pytest.fixture
def unified_session_log_path(logs_dir, config_override):
    config_override('log.root', str(logs_dir))
    config_override('log.session_subpath', 'session.log')
    config_override('log.unified_session_log', True)
    return logs_dir.join('session.log')


@pytest.fixture(autouse=True)
def override_log_level(config_override):
    config_override('log.core_log_level', logbook.TRACE)


@pytest.fixture
def session():
    session = run_tests_assert_success(SampleTest)
    return session


@pytest.fixture
def errors_log_path(request, config_override, tmpdir, logs_dir):
    subpath = 'subdir/errors.log'
    config_override('log.highlights_subpath', subpath)
    return logs_dir.join('files').join(subpath)


@pytest.fixture
def links_dir(logs_dir):
    return logs_dir.join("links")


@pytest.fixture
def files_dir(logs_dir):
    return logs_dir.join("files")


_TOKEN = "logging-test"
_SESSION_START_MARK = "session-start-mark"
_SESSION_END_MARK = "session-end-mark"

_silenced_logger = logbook.Logger("silenced_logger")


################################################################################
## Legacy Tests


class LogFormattingTest(TestCase):

    def setUp(self):
        super(LogFormattingTest, self).setUp()
        self.log_path = self.get_new_path()
        self.override_config(
            "log.root", self.log_path
        )
        self.override_config(
            "log.format", "-- {record.message} --"
        )
        self.override_config("log.subpath", "debug.log")

    def test(self):
        self.session = run_tests_assert_success(SampleTest)
        with open(os.path.join(self.log_path, "debug.log")) as logfile:
            for line in logfile:
                self.assertTrue(line.startswith("-- "))
                self.assertTrue(line.endswith(" --\n"))


class LoggingTest(TestCase):

    def test(self):
        self.log_path = self.get_new_path()
        self.override_config(
            "log.root",
            self.log_path,
        )
        self.override_config(
            "log.subpath",
            os.path.join("{context.session.id}",
                         "{context.test.__slash__.test_index0:03}-{context.test_id}", "debug.log")
        )
        self.override_config(
            "log.session_subpath",
            os.path.join("{context.session.id}", "debug.log")
        )
        self.override_config(
            "log.silence_loggers",
            [_silenced_logger.name]
        )

        self.addCleanup(gossip.unregister_token, _TOKEN)
        slash.hooks.session_start.register(  # pylint: disable=no-member
            functools.partial(_mark, _SESSION_START_MARK), token=_TOKEN)

        slash.hooks.session_end.register(  # pylint: disable=no-member
            functools.partial(_mark, _SESSION_END_MARK), token=_TOKEN)
        self.addCleanup(gossip.unregister_token, _TOKEN)

        self.session = run_tests_assert_success(SampleTest)
        self.tests_metadata = [
            result.test_metadata for result in self.session.results.iter_test_results()]
        self._test_all_run()
        self._test_test_logs_written()
        self._test_session_logs()
        self._test_no_silenced_logger_records()

    def _test_all_run(self):
        methods = [
            method_name for method_name in dir(SampleTest)
            if method_name.startswith("test")
        ]
        self.assertTrue(methods)
        self.assertEqual(len(self.tests_metadata), len(methods))

    def _test_test_logs_written(self):
        for test_metadata in self.tests_metadata:
            test_dir = "{:03}-{}".format(test_metadata.test_index0, test_metadata.id)
            log_path = os.path.join(
                self.log_path, self.session.id, test_dir, "debug.log")
            with open(log_path) as f:
                data = f.read()
            for other_test in self.tests_metadata:
                if other_test.id != test_metadata.id:
                    self.assertNotIn(other_test.id, data)
            self.assertNotIn(_SESSION_START_MARK, data)
            self.assertNotIn(_SESSION_END_MARK, data)

    def _test_session_logs(self):
        with open(os.path.join(self.log_path, self.session.id, "debug.log")) as f:
            data = f.read()
        self.assertIn(_SESSION_START_MARK, data)
        self.assertIn(_SESSION_END_MARK, data)
        for test_id in (t.id for t in self.tests_metadata):
            self.assertNotIn(test_id, data)

    def _test_no_silenced_logger_records(self):
        for path, _, filenames in os.walk(self.log_path):
            for filename in filenames:
                assert filename.endswith(".log")
                filename = os.path.join(path, filename)
                with open(filename) as f:
                    assert _silenced_logger.name not in f.read(
                    ), "Silenced logs appear in log file {}".format(filename)


class ExtraLoggersTest(TestCase):

    def setUp(self):
        super(ExtraLoggersTest, self).setUp()
        self.session = slash.Session()
        self.handler = logbook.TestHandler()
        self.addCleanup(slash.log.remove_all_extra_handlers)
        slash.log.add_log_handler(self.handler)

    def test(self):
        with self.session:
            run_tests_in_session(SampleTest, session=self.session)
        for test_result in self.session.results.iter_test_results():
            for record in self.handler.records:
                if test_result.test_id in record.message:
                    break
            else:
                self.fail(
                    "Test id {} does not appear in logger".format(test_result.test_id))


class SampleTest(slash.Test):

    def test_1(self):
        _mark()

    def test_2(self):
        _silenced_logger.error("error")
        _silenced_logger.info("info")
        _silenced_logger.debug("debug")
        _mark()


def _mark(text=None):
    if text is None:
        text = slash.context.test_id
    slash.logger.debug(text)


class TestLocaltimeLogging(TestCase):

    def setUp(self):
        super(TestLocaltimeLogging, self).setUp()
        self.assertFalse(slash.config.root.log.localtime)
        self.path = self.get_new_path()
        self.override_config(
            "log.localtime", True)
        self.override_config(
            "log.root", self.path)

    def test_local_time(self):
        with slash.Session() as session:  # pylint: disable=unused-variable
            slash.logger.info("Hello")
        self.assertNotEqual(os.listdir(self.path), [])



# pylint: disable=redefined-outer-name

import gossip
import pytest
import slash
import slash.core.requirements
from .utils import make_runnable_tests
from .utils.suite_writer.suite import Suite
from .utils.code_formatter import CodeFormatter
from slash._compat import ExitStack

_UNMET_REQ_DECORATOR = 'slash.requires(lambda: False)'
_MET_REQ_DECORATOR = 'slash.requires(lambda: True)'

def test_requirements_mismatch_session_success(suite, suite_test):
    suite_test.add_decorator('slash.requires(False)')
    suite_test.expect_skip()
    summary = suite.run()
    assert summary.session.results.is_success(allow_skips=True)


@pytest.mark.parametrize('requirement_fullfilled', [True, False])
@pytest.mark.parametrize('use_message', [True, False])
@pytest.mark.parametrize('use_fixtures', [True, False])
@pytest.mark.parametrize('message_in_retval', [True, False])
def test_requirements(suite, suite_test, requirement_fullfilled, use_fixtures, use_message, message_in_retval):

    message = "requires something very important"
    if use_message and message_in_retval:
        retval = '({}, {!r})'.format(requirement_fullfilled, message)
    else:
        retval = requirement_fullfilled

    suite_test.add_decorator('slash.requires((lambda: {}), {!r})'.format(retval, message if use_message and not message_in_retval else ''))
    if not requirement_fullfilled:
        suite_test.expect_skip()

    if use_fixtures:
        suite_test.depend_on_fixture(
            suite.slashconf.add_fixture())
    results = suite.run()
    if requirement_fullfilled:
        assert results[suite_test].is_success()
    else:
        assert not results[suite_test].is_started()
        assert results[suite_test].is_skip()
        if use_message:
            [skip] = results[suite_test].get_skips()
            assert message in skip


def test_requirements_functions_no_message(suite, suite_test):
    suite_test.add_decorator(_UNMET_REQ_DECORATOR)
    suite_test.expect_skip()
    results = suite.run()
    result = results[suite_test]
    [skip] = result.get_skips()
    assert 'lambda' in skip


def test_requirements_on_class():

    def req1():
        pass

    def req2():
        pass

    @slash.requires(req1)
    class Test(slash.Test):

        @slash.requires(req2)
        def test_something(self):
            pass

    with slash.Session():
        [test] = make_runnable_tests(Test)  # pylint: disable=unbalanced-tuple-unpacking


    assert set([r._req for r in test.get_requirements()]) == set([req1, req2])  # pylint: disable=protected-access


@pytest.fixture
def filename_test_fixture(tmpdir):
    returned = str(tmpdir.join('testfile.py'))

    with open(returned, 'w') as f:
        with ExitStack() as stack:
            code = CodeFormatter(f)

            code.writeln('import slash')
            code.writeln('@slash.fixture')
            code.writeln('@slash.requires({}, {})'.format(_UNMET_REQ_DECORATOR, '"msg1"'))
            code.writeln('def fixture():')
            with code.indented():
                code.writeln('return 1')

            code.writeln('@slash.fixture(autouse=True)')
            code.writeln('@slash.requires({}, {})'.format(_MET_REQ_DECORATOR, '"msg2"'))
            code.writeln('def fixture1():')
            with code.indented():
                code.writeln('return 1')

            code.writeln('class Test(slash.Test):')
            stack.enter_context(code.indented())

            code.write('def test_1(')
            code.write('self, ')
            code.writeln('fixture):')
            with code.indented():
                code.writeln('pass')
    return returned


def test_requirements_on_class_with_fixture_and_autouse_fixture(filename_test_fixture):
    with slash.Session():
        [test] = make_runnable_tests(filename_test_fixture)  # pylint: disable=unbalanced-tuple-unpacking
    assert sorted([str(r) for r in test.get_requirements()]) == ['msg1', 'msg2']


def test_unmet_requirements_trigger_avoided_test_hook(suite, suite_test):

    suite_test.add_decorator(_UNMET_REQ_DECORATOR)
    suite_test.expect_skip()


    @gossip.register('slash.test_avoided')
    def test_avoided(reason):  # pylint: disable=unused-variable
        slash.context.result.data['avoided'] = {'reason': reason,
                                                'test_name': slash.context.test.__slash__.address}

    summary = suite.run()
    avoided_result = summary[suite_test]


    for r in summary.session.results.iter_all_results():
        if r is avoided_result:
            assert 'avoided' in r.data
            assert 'lambda' in r.data['avoided']['reason']
            assert 'unmet requirement' in r.data['avoided']['reason'].lower()
            assert r.data['avoided']['test_name'].split('_')[-1] == suite_test.id
        else:
            assert 'avoided' not in r.data


def test_adding_requirement_objects():

    class MyRequirement(slash.core.requirements.Requirement):
        pass

    req = MyRequirement('bla')

    @slash.requires(req)
    def test_something():
        pass

    with slash.Session():
        [test] = make_runnable_tests(test_something) # pylint: disable=unbalanced-tuple-unpacking

    reqs = test.get_requirements()
    assert len(reqs) == 1 and reqs[0] is req


def test_cannot_specify_message_with_requirement_object():

    class MyRequirement(slash.core.requirements.Requirement):
        pass

    with pytest.raises(AssertionError) as caught:
        slash.requires(MyRequirement(''), 'message')

    assert 'specify message' in str(caught.value)

@pytest.mark.parametrize('is_fixture_requirement_unmet', [True, False])
def test_fixture_and_test_requirements(suite, suite_test, is_fixture_requirement_unmet):
    suite_test.depend_on_fixture(suite.slashconf.add_fixture())
    if is_fixture_requirement_unmet:
        suite_test._fixtures[0][1].add_decorator(_UNMET_REQ_DECORATOR) # pylint: disable=protected-access
        suite_test.add_decorator(_MET_REQ_DECORATOR)
    else:
        suite_test._fixtures[0][1].add_decorator(_MET_REQ_DECORATOR) # pylint: disable=protected-access
        suite_test.add_decorator(_UNMET_REQ_DECORATOR)

    suite_test.expect_skip()
    results = suite.run()
    assert results[suite_test].is_skip()
    result = results[suite_test]
    [skip] = result.get_skips()
    assert 'lambda' in skip

def test_fixture_of_fixture_requirement(suite, suite_test):
    suite_test.add_decorator(_UNMET_REQ_DECORATOR)
    suite_test.depend_on_fixture(suite.slashconf.add_fixture())
    suite_test._fixtures[0][1].add_decorator(_MET_REQ_DECORATOR) # pylint: disable=protected-access
    suite_test.expect_skip()
    results = suite.run()
    assert results[suite_test].is_skip()

    result = results[suite_test]
    [skip] = result.get_skips()
    assert 'lambda' in skip


def test_autouse_fixture_requirement():
    suite = Suite()
    for _ in range(5):
        test = suite.add_test(type='function')
        test.expect_skip()
    fixture = suite.get_last_file().add_fixture(autouse=True)
    fixture.add_decorator(_UNMET_REQ_DECORATOR)
    suite.run()



from ..utils.marks import mark
from ..utils import parallel_utils
from .interface import PluginInterface # pylint: disable=unused-import



def registers_on(hook_name, **kwargs):
    """Marks the decorated plugin method to register on a custom hook, rather than
    the method name in the 'slash' group, which is the default behavior for plugins

    Specifying ``registers_on(None)`` means that this is not a hook entry point at all.

    .. note:: All keyword arguments are forwarded to gossip's ``register`` API
    """
    return mark("register_on", RegistrationInfo(hook_name, expect_exists=False, register_kwargs=kwargs), append=True)


def parallel_mode(mode):
    """Marks compatibility of a specific plugin to parallel execution.

    :param mode: Can be either ``disabled``, ``enabled``, ``parent-only`` or ``child-only``
    """
    possible_values = parallel_utils.ParallelPluginModes.MODES
    assert mode in possible_values, "parallel mode value must be one of {}".format(possible_values)
    return mark("parallel_mode", mode)


def register_if(condition):
    """Marks the decorated plugins method to only be registered if *condition* is ``True``
    """
    return mark("register_if", condition)


def active(plugin_class):
    """Decorator for automatically installing and activating a plugin upon definition
    """
    plugin = plugin_class()
    manager.install(plugin)
    manager.activate(plugin)

    return plugin_class


def needs(what):
    return mark("plugin_needs", what, append=True)


def provides(what):
    return mark("plugin_provides", what, append=True)


from .plugin_manager import manager, IncompatiblePlugin, UnknownPlugin, IllegalPluginName, RegistrationInfo # pylint: disable=unused-import



import copy
from .tagging import NO_TAGS

class RunnableTest(object):

    """
    This class is meant to serve as a base class to any test that can
    actually be executed by the Slash runner.
    """
    __slash__ = None

    def __init__(self, fixture_store, fixture_namespace, variation):
        super(RunnableTest, self).__init__()
        self._fixture_store = fixture_store
        self._fixture_namespace = fixture_namespace
        self._variation = variation

    def clone(self):
        returned = copy.copy(self)
        for copied_attr in ('_variation', '__slash__'):
            setattr(returned, copied_attr, copy.copy(getattr(returned, copied_attr)))
        return returned

    @property
    def id(self):
        return self.__slash__.id

    def get_variation(self):
        return self._variation

    def get_address_in_factory(self):
        raise NotImplementedError() # pragma: no cover

    def run(self):
        """
        This method is meant to be overriden by derived classes to actually
        perform the test logic
        """
        raise NotImplementedError()  # pragma: no cover

    def get_tags(self):
        return NO_TAGS

    def get_test_function(self):
        raise NotImplementedError() # pragma: no cover

    def get_unmet_requirements(self):
        returned = []
        for req in self.get_requirements():
            is_met, reason = req.is_met()
            if not is_met:
                returned.append((req, reason))
        return returned

    def _get_fixtures_requirements(self):
        fixture_requirements = [item for fixture in self.get_required_fixture_objects() for item in fixture.get_requirements(self._fixture_store)]
        autouse_fixture_requirements = [item for fixture in self._fixture_store.iter_autouse_fixtures_in_namespace(self.get_fixture_namespace()) \
                            for item in fixture.get_requirements(self._fixture_store)]
        return list(set(fixture_requirements + autouse_fixture_requirements))

    def get_requirements(self):
        raise NotImplementedError() # pragma: no cover

    def get_required_fixture_objects(self):
        raise NotImplementedError() # pragma: no cover

    def get_fixture_namespace(self):
        return self._fixture_namespace

    def __repr__(self):
        return '<Runnable test {!r}>'.format(self.__slash__)



import pytest


def test_set_test_detail(result):
    result.set_test_detail('x', 'y')
    assert result.details.all() == {'x': 'y'}


def test_set(result):
    result.details.set('x', 'y')
    assert result.details.all() == {'x': 'y'}


def test_append(result):
    result.details.append('x', 'y')
    assert result.details.all() == {'x': ['y']}


def test_append_invalid_type(result):
    result.details.set('x', 'y')
    with pytest.raises(TypeError):
        result.details.append('x', 'y')
    assert result.details.all() == {'x': 'y'}

def test_bool(result):
    assert not result.details
    assert not bool(result.details)
    result.details.set('x', 'y')
    assert result.details
    assert bool(result.details)

def test_contains(result):
    assert 'x' not in result.details
    result.details.set('x', 'y')
    assert 'x' in result.details



import pytest
from .utils import skip_on_py2, skip_on_pypy

@skip_on_py2
@skip_on_pypy
@pytest.mark.parametrize('alias_with_attribute', [True, False])
def test_fixture_annotations(suite, suite_test, get_fixture_location, alias_with_attribute):

    fixture = get_fixture_location(suite_test).add_fixture()
    suite_test.depend_on_fixture(fixture, alias=True, alias_with_attribute=alias_with_attribute)

    suite.run()

@skip_on_py2
@skip_on_pypy
@pytest.mark.parametrize('alias_with_attribute', [True, False])
def test_nested_fixture_annotations(suite, suite_test, get_fixture_location, alias_with_attribute):

    fixture = get_fixture_location(suite_test).add_fixture()
    fixture2 = get_fixture_location(suite_test).add_fixture()
    fixture.depend_on_fixture(fixture2, alias=True, alias_with_attribute=alias_with_attribute)
    suite_test.depend_on_fixture(fixture)

    suite.run()



import argparse
import sys

import colorama
from slash.utils.cli_utils import make_styler, Printer

from slash import site
from slash.plugins import manager

_title_style = make_styler(colorama.Fore.WHITE + colorama.Style.BRIGHT)  # pylint: disable=no-member
_disabled_style = make_styler(colorama.Fore.BLACK + colorama.Style.BRIGHT)  # pylint: disable=no-member
_enabled_style = make_styler(colorama.Fore.GREEN)  # pylint: disable=no-member

_link_style = make_styler(colorama.Fore.CYAN)  # pylint: disable=no-member


def _get_parser():
    parser = argparse.ArgumentParser('slash list-plugins [options]')
    parser.add_argument('--force-color', dest='force_color', action='store_true', default=False)
    parser.add_argument('--no-color', dest='enable_color', action='store_false', default=True)
    return parser


def slash_list_plugins(args, report_stream=sys.stdout):
    parser = _get_parser()
    parsed_args = parser.parse_args(args)

    _print = Printer(report_stream, force_color=parsed_args.force_color, enable_color=parsed_args.enable_color)

    site.load()

    active = manager.get_future_active_plugins()

    for plugin in sorted(manager.get_installed_plugins(include_internals=False).values(), key=lambda p: p.get_name()):
        name = plugin.get_name()
        _print(_title_style(name), end=' ')
        if name in active:
            _print(_enabled_style('active (use --without-{} to deactivate'.format(name)))
        else:
            _print(_disabled_style('inactive (use --with-{} to activate)'.format(name)))
        if plugin.__doc__:
            for line in plugin.__doc__.splitlines():
                if line.strip():
                    _print('\t', line.strip())

    return 0



from ..conf import config

def is_parallel_session():
    return config.root.parallel.num_workers != 0


def is_parent_session():
    return is_parallel_session() and config.root.parallel.worker_id is None


def is_child_session():
    return config.root.parallel.worker_id is not None


class ParallelPluginModes(object):
    DISABLED, ENABLED, PARENT_ONLY, CHILD_ONLY = MODES = (
        'disabled', 'enabled', 'parent-only', 'child-only'
    )



import sys

import emport

from slash.core.error import Error
from slash.utils.traceback_utils import _MAX_VARIABLE_VALUE_LENGTH



def test_traceback_line_numbers(tmpdir):
    filename = tmpdir.join('filename.py')

    with filename.open('w') as f:
        f.write('''from contextlib import contextmanager
def f():
    with context():
        a = 1
        b = 2
        g()
        c = 3
        d = 4

def g():
    1/0

@contextmanager
def context():
    yield
''')

    mod = emport.import_file(str(filename))
    try:
        mod.f()
    except ZeroDivisionError:
        err = Error(exc_info=sys.exc_info())
    else:
        assert False, 'did not fail'

    assert err.traceback.frames[-2].lineno == 6


def test_variable_capping():

    def f():
        g()

    def g():
        long_var = 'a' * 1000
        assert len(long_var) > _MAX_VARIABLE_VALUE_LENGTH
        1/0                     # pylint: disable=pointless-statement

    try:
        f()
    except ZeroDivisionError:
        err = Error(exc_info=sys.exc_info())

    distilled = err.traceback.to_list()
    assert len(distilled[-1]['locals']['long_var']['value']) == _MAX_VARIABLE_VALUE_LENGTH




def test_is_test_code(suite, suite_test):
    suite_test.when_run.error()
    summary = suite.run()
    [result] = summary.get_all_results_for_test(suite_test)
    [err] = result.get_errors()
    assert err.traceback.frames[-1].is_in_test_code()

    error_json = err.traceback.to_list()
    assert error_json[-1]['is_in_test_code']


def test_self_attribute_throws():

    class CustomException(Exception):
        pass

    def func():
        x = DangerousObject()
        x.method()

    class DangerousObject(object):

        def __getattribute__(self, attr):
            if attr == '__dict__':
                1/0  # pylint: disable=pointless-statement
            return super(DangerousObject, self).__getattribute__(attr)

        def method(self):
            raise CustomException()

    try:
        func()
    except CustomException:
        error = Error(exc_info=sys.exc_info())
    else:
        assert False, 'Did not raise'

    locals = error.traceback.frames[-1].locals
    assert 'self' in locals
    for key in locals:
        assert 'self.' not in key



from docutils.parsers.rst import directives, Directive
from docutils import nodes
from slash.conf import config


class ConfigDoc(Directive):
    required_arguments = 0
    optional_arguments = 0

    def run(self):
        returned = []
        for path, leaf in config.traverse_leaves():
            section = nodes.section(names=["conf." + path])
            self.state.document.note_explicit_target(section)
            section.append(nodes.title(text=path))
            section.append(nodes.strong(text="Default: {}".format(leaf.get_value())))
            if leaf.metadata and "doc" in leaf.metadata:
                section.append(nodes.paragraph(text=str(leaf.metadata["doc"])))
            returned.append(section)
        return returned


def setup(app):                 # pylint: disable=unused-argument
    directives.register_directive('config_doc', ConfigDoc)



from slash._compat import izip_longest
from uuid import uuid4

from .element import Element


class Parameter(Element):

    def __init__(self, suite, num_values=3, values=None):
        super(Parameter, self).__init__(suite)
        self.name = 'param_{}'.format(self.id)
        if values is None:
            values = [str(uuid4()) for _ in range(num_values)]
        self.values = values
        self.labels = []

    def write_decorator(self, code_formatter):
        code_formatter.writeln('@slash.parametrize({!r}, {})'.format(
            self.name, self._format_values()))

    def _format_values(self):
        returned = '['
        for label, value in izip_longest(self.labels, self.values):
            assert value is not None
            if label is None:
                returned += repr(value)
            else:
                returned += 'slash.param({!r}, {!r})'.format(label, value)
            returned += ', '
        returned += ']'
        return returned


    def add_labels(self):
        assert not self.labels
        self.labels = ['label_{}'.format(str(uuid4()).replace('-', '_')[:20]) for value in self.values]



from __future__ import print_function
import functools
import os
import re

from slash.frontend.slash_list import slash_list
from slash._compat import StringIO
from .utils.suite_writer import Suite

import pytest


@pytest.mark.parametrize('flag', ["--only-fixtures", "--only-tests", None])
def test_slash_list(suite, flag):
    suite.debug_info = False
    _ = suite.slashconf.add_fixture()
    path = suite.commit()
    report_stream = StringIO()

    args = [path]
    if flag is not None:
        args.append(flag)
    slash_list(args, report_stream)
    assert report_stream.getvalue()


@pytest.mark.parametrize('allow_empty', [True, False])
def test_slash_list_without_any_tests(allow_empty):
    empty_suite = Suite()
    empty_suite.debug_info = False
    path = empty_suite.commit()
    report_stream = StringIO()
    args = [path]
    if allow_empty:
        args.append('--allow-empty')
    rc = slash_list(args, report_stream)
    expected_rc = 0 if allow_empty else 1
    assert rc == expected_rc


def test_slash_list_tests_without_tags(suite):
    suite.debug_info = False
    path = suite.commit()
    report_stream = StringIO()
    args = [path, '--show-tags', '--no-output']
    slash_list(args, report_stream)
    output = report_stream.getvalue()
    assert not output


@pytest.mark.parametrize('should_show_tags', [True, False])
def test_slash_list_tests_with_or_without_tags(suite, should_show_tags, suite_test):
    suite_test.add_decorator('slash.tag("bla")')
    suite.debug_info = False
    path = suite.commit()
    report_stream = StringIO()
    args = [path, '--only-tests']
    if should_show_tags:
        args.append('--show-tags')
    slash_list(args, report_stream)
    output = report_stream.getvalue()
    assert ('Tags' in output) == (should_show_tags)


@pytest.mark.parametrize('relative', [True, False])
def test_slash_list_tests_relative_or_not(suite, relative):
    suite.debug_info = False
    path = suite.commit()
    report_stream = StringIO()
    args = [path]
    if relative:
        args.append('--relative-paths')
    slash_list(args, report_stream)
    output_lines = {
        _strip(line)
        for line in report_stream.getvalue().splitlines()
    } - {"Tests", "Fixtures"}
    assert output_lines
    for filename in output_lines:
        assert os.path.isabs(filename) == (not relative)


@pytest.mark.parametrize('invalid', [None, 'test', 'method'])
@pytest.mark.parametrize('allow_empty', [True, False])
def test_slash_list_suite_file_incorrect_names(tmpdir, invalid, allow_empty):

    with tmpdir.join('test_file.py').open('w') as f:
        _print = functools.partial(print, file=f)

        _print('import slash')
        _print('class TestSomething(slash.Test):')
        _print('    def test_method(self):')
        _print('        pass')
        _print()
        _print('def test_function():')
        _print('    pass')

    function_name_remainder = method_name_remainder = ''

    if invalid == 'test':
        function_name_remainder = 'f'
    elif invalid == 'method':
        method_name_remainder = 'm'
    elif invalid is not None:
        raise NotImplementedError() # pragma: no cover

    error_stream = StringIO()

    with tmpdir.join('suitefile').open('w') as suite_file:
        _print = functools.partial(print, file=suite_file)
        _print('{}:TestSomething.test_method{}'.format(f.name, method_name_remainder))
        _print('{}:test_function{}'.format(f.name, function_name_remainder))

    args = ['-f', suite_file.name]
    if allow_empty:
        args.append('--allow-empty') # make sure allow-empty does not affect invalid name lookup
    result = slash_list(args, error_stream=error_stream)

    if invalid is None:
        assert result == 0
    else:
        assert result != 0
        assert 'Could not load tests' in error_stream.getvalue()
        if invalid == 'test':
            assert 'test_functionf' in error_stream.getvalue()
        elif invalid == 'method':
            assert 'test_methodm' in error_stream.getvalue()
        else:
            raise NotImplementedError() # pragma: no cover


@pytest.mark.parametrize('no_tests', [True, False])
def test_slash_list_with_warnings(tmpdir, no_tests):
    with tmpdir.join('test_file.py').open('w') as fp:
        _print = functools.partial(print, file=fp)
        _print('import warnings')
        _print('from slash import hooks')

        _print()
        _print('warnings.warn("Some warning")')
        _print('@hooks.after_session_start.register')
        _print('def _warn():')
        _print('    warnings.warn("Some warning")')

        if not no_tests:
            _print()
            _print('def test_me():')
            _print('    pass')

    error_stream = StringIO()
    args = [fp.name, '--warnings-as-errors']
    if no_tests:
        args.append('--allow-empty')
    result = slash_list(args, error_stream=error_stream)

    errors = error_stream.getvalue()
    assert 'Could not load tests' not in errors
    assert result != 0


def _strip(line):
    return re.sub(r'\x1b\[.+?m', '', line).strip()



import functools
import itertools
import os
import pickle
import sys
from numbers import Number

import gossip
import logbook
from vintage import deprecated

from .. import hooks
from .._compat import OrderedDict, itervalues
from ..ctx import context
from ..exception_handling import capture_sentry_exception, handling_exceptions
from .. import exceptions
from ..utils.exception_mark import ExceptionMarker
from ..utils.interactive import notify_if_slow_context
from ..utils.python import unpickle
from .details import Details
from .error import Error

_logger = logbook.Logger(__name__)

_ADDED_TO_RESULT = ExceptionMarker('added_to_result')


class Result(object):
    """Represents a single result for a test which was run
    """
    pickle_key_blacklist = {'test_metadata', 'facts'}

    def __init__(self, test_metadata=None):
        super(Result, self).__init__()
        self.test_metadata = test_metadata
        #: dictionary to be use by tests and plugins to store result-related information for later analysis
        self.data = {}
        self._errors = []
        self._failures = []
        self._skips = []
        #: a :class:`slash.core.details.Details` instance for storing additional test details
        self.details = Details()
        self.facts = Details(set_callback=self._fact_set_callback)
        self._started = False
        self._finished = False
        self._interrupted = False
        self._log_path = None
        self._extra_logs = []

    def _fact_set_callback(self, fact_name, fact_value):
        gossip.trigger('slash.fact_set', name=fact_name, value=fact_value)

    def is_global_result(self):
        return False

    def serialize(self):
        serialized_object = {}
        for key, value in vars(self).items():
            if key not in Result.pickle_key_blacklist:
                try:
                    serialized_object[key] = pickle.dumps(value)
                except (pickle.PicklingError, TypeError):
                    _logger.debug('Failed serializing result, skipping this value. key = {}', key)
                    capture_sentry_exception()
        return serialized_object

    def deserialize(self, result_dict):
        for key, value in result_dict.items():
            try:
                self.__dict__[key] = unpickle(value)
            except TypeError:
                _logger.error('Error when deserialize reult, skipping this value. key = {}', key)
        for failure in self._failures:
            self.add_failure(failure, append=False)
        for error in self._errors:
            self.add_error(error, append=False)
        for skip in self._skips:
            self.add_skip(skip, append=False)

    def add_exception(self, exc_info=None):
        """Adds the currently active exception, assuming it wasn't already added to a result
        """
        if exc_info is None:
            exc_info = sys.exc_info()
        _, exc_value, _ = exc_info  # pylint: disable=unpacking-non-sequence

        if _ADDED_TO_RESULT.is_exception_marked(exc_value):
            return

        _ADDED_TO_RESULT.mark_exception(exc_value)
        if isinstance(exc_value, exceptions.FAILURE_EXCEPTION_TYPES):
            self.add_failure()
        elif isinstance(exc_value, context.session.get_skip_exception_types()):
            self.add_skip(getattr(exc_value, 'reason', str(exc_value)))
        elif isinstance(exc_value, exceptions.INTERRUPTION_EXCEPTIONS):
            err = Error.capture_exception(exc_info=exc_info)
            hooks.interruption_added(result=self, exception=err) # pylint: disable=no-member
            session_result = context.session.results.global_result
            interrupted_test = self.is_interrupted()
            interrupted_session = session_result.is_interrupted()
            if not self.is_global_result():
                self.mark_interrupted()
                if not interrupted_test and not context.session.has_children():
                    with notify_if_slow_context(message="Cleaning up test due to interrupt. Please wait..."),\
                         handling_exceptions(swallow=True):

                        hooks.test_interrupt() # pylint: disable=no-member
            if not interrupted_session:
                session_result.mark_interrupted()

        elif not isinstance(exc_value, GeneratorExit):
            #skip keyboardinterrupt and system exit
            self.add_error(exc_info=exc_info)
        else:
            _logger.trace('Ignoring GeneratorExit exception')

    def has_errors_or_failures(self):
        return bool(self._failures or self._errors)

    def get_log_path(self):
        """Returns log path
        """
        return self._log_path

    def set_log_path(self, path):
        """Set log path
        """
        self._log_path = path

    def get_log_dir(self):
        """Returns log's directory.
        """
        if self._log_path is None:
            return None
        return os.path.dirname(self._log_path)

    def add_extra_log_path(self, path):
        """Add additional log path. This path will be added to the list returns by get_log_paths
        """
        self._extra_logs.append(path)

    def get_log_paths(self):
        """Returns a list of all log paths
        """
        logs = []
        if self._log_path:
            logs.append(self._log_path)
        return logs + list(self._extra_logs)

    def is_started(self):
        return self._started

    def is_not_run(self):
        return not self._started

    def mark_started(self):
        self._started = True

    def is_error(self):
        return bool(self._errors)

    @property
    def test_id(self):
        return self.test_metadata.id

    def is_failure(self):
        return bool(self._failures)

    def is_just_failure(self):
        """Indicates this is a pure failure, without errors involved"""
        return self.is_failure() and not self.is_error()

    def is_skip(self):
        return bool(self._skips)

    def is_run_and_skip(self):
        return self.is_started() and self.is_skip()

    def is_success(self, allow_skips=False):
        if not self.is_started():
            return allow_skips
        returned = not self._errors and not self._failures and not self._interrupted
        if not allow_skips:
            returned &= not self._skips
        return returned

    def is_success_finished(self):
        return self.is_success() and self.is_finished()

    def is_finished(self):
        return self._finished

    def mark_finished(self):
        self._finished = True

    def mark_interrupted(self):
        self._interrupted = True

    def is_interrupted(self):
        return self._interrupted

    def add_error(self, e=None, frame_correction=0, exc_info=None, append=True):
        """Adds a failure to the result
        """
        err = self._add_error(self._errors, e, frame_correction=frame_correction + 1, exc_info=exc_info, append=append)
        context.reporter.report_test_error_added(context.test, err)
        return err

    def add_failure(self, e=None, frame_correction=0, exc_info=None, append=True):
        """Adds a failure to the result
        """
        err = self._add_error(self._failures, e, frame_correction=frame_correction + 1, exc_info=exc_info, is_failure=True, append=append)
        context.reporter.report_test_failure_added(context.test, err)
        return err

    def set_test_detail(self, key, value):
        """Adds a generic detail to this test result, which can be later inspected or used
        """
        self.details.set(key, value)

    def _add_error(self, error_list, error=None, frame_correction=0, exc_info=None, is_failure=False, append=True):
        try:
            if error is None:
                error = Error.capture_exception(exc_info=exc_info)
                if error is None:
                    raise RuntimeError('add_error() must be called with either an argument or in an active exception')
            if not isinstance(error, Error):
                error = Error(error, frame_correction=frame_correction + 1, exc_info=exc_info)
            if is_failure:
                # force the error object to be marked as failure
                error.mark_as_failure()
            error.log_added()
            if append:
                error_list.append(error)
            #report children errors or parent session errors only
            if not context.session or not context.session.has_children() or self.is_global_result():
                hooks.error_added(result=self, error=error)  # pylint: disable=no-member
            error.forget_exc_info()
            return error
        except Exception:
            _logger.error("Failed to add error to result", exc_info=True)
            raise

    def add_skip(self, reason, append=True):
        if append:
            self._skips.append(reason)
        context.reporter.report_test_skip_added(context.test, reason)

    def get_errors(self):
        """Returns the list of errors recorded for this result

        :return: a list of :class:`slash.core.error.Error` objects
        """
        return self._errors

    def get_failures(self):
        """Returns the list of failures recorded for this result

        :return: a list of :class:`slash.core.error.Error` objects
        """
        return self._failures

    @deprecated('Use result.details.all()', since='0.20.0')
    def get_additional_details(self):
        return self.details.all()

    def get_skips(self):
        return self._skips

    def has_skips(self):
        return bool(self._skips)

    def has_fatal_errors(self):
        return any(e.is_fatal() for e in
                   itertools.chain(self._errors, self._failures))

    has_fatal_exception = has_fatal_errors

    def __repr__(self):
        return "<{} ({})>".format(
            self.__class__.__name__,
            ", ".join(
                attr
                for attr in ("success", "error", "failure", "skip", "finished", "interrupted")
                if getattr(self, "is_{}".format(attr))()
            )
        )


class GlobalResult(Result):

    def __init__(self, session_results=None):
        super(GlobalResult, self).__init__()
        self._session_results = session_results

    def is_global_result(self):
        return True

    def is_success(self, allow_skips=False):
        if not super(GlobalResult, self).is_success(allow_skips=allow_skips):
            return False
        if self._session_results is None:
            return True
        if self._session_results.session.has_children() and self._session_results.session.parallel_manager.server.worker_session_error_reported:
            return False
        return all(result.is_success(allow_skips=allow_skips) for result in self._session_results.iter_test_results())


class SessionResults(object):

    def __init__(self, session):
        super(SessionResults, self).__init__()
        self.session = session
        self.global_result = GlobalResult(self)
        self._results_dict = OrderedDict()
        self._iterator = functools.partial(itervalues, self._results_dict)

    def __len__(self):
        return len(self._results_dict)

    def __repr__(self):
        return '<Results: {} successful, {} errors, {} failures, {} skips>'.format(
            self.get_num_successful(),
            self.get_num_errors(),
            self.get_num_failures(),
            self.get_num_skipped())

    def iter_all_additional_details(self):
        for result in self.iter_all_results():
            if result.details:
                yield result, result.get_additional_details()

    def iter_all_failures(self):
        """Iterates over all results which have failures

        yields tuples of the form (result, failure_list)
        """
        for result in self.iter_all_results():
            if result.get_failures():
                yield result, result.get_failures()

    def iter_all_errors(self):
        """Iterates over all results which have errors

        yields tuples of the form (result, errors_list)
        """
        for result in self.iter_all_results():
            if result.get_errors():
                yield result, result.get_errors()

    @property
    def current(self):
        """Obtains the currently running result, if exists

        Otherwise, returns the global result object
        """
        return context.result

    def __iter__(self):
        return self._iterator()

    def is_success(self, allow_skips=False):
        """Indicates whether this run is successful

        :param allow_skips: Whether to consider skips as unsuccessful
        """
        return self.global_result.is_success(allow_skips=allow_skips)

    def is_interrupted(self):
        """Indicates if this session was interrupted
        """
        return any(result.is_interrupted() for result in self._iterator())

    def get_num_results(self):
        return len(self._results_dict)

    def get_num_started(self):
        return self._count(Result.is_started, include_global=False)

    def get_num_successful(self):
        return self._count(Result.is_success_finished, include_global=False)

    def get_num_errors(self):
        return self._count(Result.is_error)

    def get_num_failures(self):
        return self._count(Result.is_just_failure)

    def get_num_skipped(self, include_not_run=True):
        if include_not_run:
            return self._count(Result.is_skip)
        return self._count(Result.is_run_and_skip)

    def get_num_not_run(self):
        return self._count(Result.is_not_run, include_global=False)

    def has_fatal_errors(self):
        """Indicates whether any result has an error marked as fatal (causing the session to terminate)
        """
        return bool(self._count(Result.has_fatal_errors))

    def _count(self, pred, include_global=True):
        returned = 0
        iterator = self.iter_all_results(
        ) if include_global else self.iter_test_results()
        for result in iterator:
            if pred(result):
                returned += 1
        return returned

    def iter_test_results(self):
        """Iterates over results belonging to tests
        """
        return iter(self)

    def iter_all_results(self):
        """Iterates over all results, ending with the global result object
        """
        return itertools.chain(self.iter_test_results(), [self.global_result])

    def create_result(self, test):
        assert test.__slash__.id not in self._results_dict
        returned = Result(test.__slash__)
        self._results_dict[test.__slash__.id] = returned
        return returned

    def get_result(self, test):
        """Returns the result stored belonging to a given test
        """
        if test.__slash__ is None:
            raise LookupError("Could not find result for {}".format(test))
        return self._results_dict[test.__slash__.id]

    def __getitem__(self, test):
        if isinstance(test, Number):
            try:
                return next(itertools.islice(itervalues(self._results_dict), test, test + 1))
            except StopIteration:
                raise IndexError()
        return self.get_result(test)
